;;; schmooz.scm: Program for extracting texinfo comments from Scheme.
;;; Copyright (C) 1998 Radey Shouman and Aubrey Jaffer.
;
;Permission to copy this software, to redistribute it, and to use it
;for any purpose is granted, subject to the following restrictions and
;understandings.
;
;1.  Any copy made of this software must include this copyright notice
;in full.
;
;2.  I have made no warrantee or representation that the operation of
;this software will be error-free, and I am under no obligation to
;provide any services, by way of maintenance, update, or otherwise.
;
;3.  In conjunction with products arising from the use of this
;material, there shall be no use of my name in any advertising,
;promotional, or sales literature without prior written consent in
;each case.

;;$Header: /usr/local/cvsroot/slib/schmooz.scm,v 1.2 1998/06/04 16:07:22 jaffer Exp $
;;$Name:  $

(provide 'schmooz)

;;; REPORT an error or warning
(define report
  (lambda args
    (display *scheme-source-name*)
    (display ": In function `")
    (display *procedure*)
    (display "': ")
    (newline)

    (display *derived-txi-name*)
    (display ": ")
    (display *output-line*)
    (display ": warning: ")
    (apply qreport args)))

(define qreport
  (lambda args
    (for-each (lambda (x) (write x) (display #\ )) args)
    (newline)))

(require 'common-list-functions)	;some
(require 'string-search)
(require 'fluid-let)
(require 'line-i/o)			;read-line
;;(require 'debug) (set! *qp-width* 100) (define qreport qpn)

;;; This allows us to test without generating files
(define *scheme-source* (current-input-port))
(define *scheme-source-name* "stdin")
(define *derived-txi* (current-output-port))
(define *derived-txi-name* "?")

(define *procedure* #f)
(define *output-line* 0)

(define CONTLINE -80)

;;; OUT indents and displays the arguments
(define (out indent . args)
  (cond ((>= indent 0)
	 (newline *derived-txi*)
	 (set! *output-line* (+ 1 *output-line*))
	 (do ((j indent (- j 8)))
	     ((> 8 j)
	      (do ((i j (- i 1)))
		  ((>= 0 i))
		(display #\  *derived-txi*)))
	   (display #\	 *derived-txi*))))
  (for-each (lambda (a)
	      (cond ((symbol? a)
		     (display a *derived-txi*))
		    ((string? a)
		     (display a *derived-txi*)
		     #+f
		     (cond ((string-index a #\nl)
			    (set! *output-line* (+ 1 *output-line*))
			    (report "newline in string" a))))
		    (else
		     (display a *derived-txi*))))
	    args))

(define (descmfilify file)
  (let ((sl (string-length file)))
    (cond ((< sl 4) file)
	  ((string-ci=? (substring file (- sl 4) sl) ".scm")
	   (substring file 0 (- sl 4)))
	  (else file))))

;; Substitute @ macros in string LINE.
;; Returns a list of strings, the first is the substituted version
;; of LINE, the rest are "@cindex " directives for Texinfo.
;; MACS is an alist of (macro-name . macro-value) pairs.
(define (substitute-macs line macs)
  (define (get-word i)
    (let loop ((j (+ i 1)))
      (cond ((>= j (string-length line))
	     (substring line i j))
	    ((or (char-alphabetic? (string-ref line j))
		 (char-numeric? (string-ref line j)))
	     (loop (+ j 1)))
	    (else (substring line i j)))))
  ;;Return (next-char-number . list-of-arguments)
  (define (get-args i)
    (let skip ((i i))
      (cond ((>= i (string-length line)) #f)
	    ((char-whitespace? (string-ref line i))
	     (skip (+ i 1)))
	    ((eqv? (string-ref line i) #\{)
	     (let loop-args ((i (+ i 1))
			     (args '()))
	       (let loop ((j i))
		 (if (>= j (string-length line))
		     #f;;error
		     (case (string-ref line j)
		       ((#\,)
			(loop-args (+ j 1)
				   (cons (substring line i j) args)))
		       ((#\}) 
			(cons (+ j 1)
			      (reverse (cons (substring line i j) args))))
		       (else (loop (+ j 1))))))))
	    (else #f))))
  (let loop ((istrt 0)
	     (i 0)
	     (res '())
	     (idxs '()))
    (cond ((>= i (string-length line))
	   (cons (apply string-append 
			(reverse 
			 (cons (substring line istrt (string-length line))
			       res)))
		 idxs))
	  ((char=? #\@ (string-ref line i))
	   (let* ((w (get-word i))
		  (symw (string->symbol w)))
	     (cond ((eq? '@cname symw)
		    (let ((args (get-args (+ i (string-length w)))))
		      (cond ((and args (= 2 (length args)))
			     (loop (car args) (car args)
				   (cons
				    (string-append
				     "@code{" (cadr args) "}")
				    (cons (substring line istrt i) res))
				   idxs))
			    (else
			     (report "@cname wrong number of args" line)
			     (loop istrt (+ i (string-length w)) res idxs)))))
		   ((eq? '@dfn symw)
		    (let* ((args (get-args (+ i (string-length w))))
			   (inxt (car args))
			   (args (cdr args)))
		      (loop inxt inxt
			    (cons (substring line istrt inxt) res)
			    (cons (string-append "@cindex " (car args))
				  idxs))))
		   ((assq symw macs) =>
		    (lambda (s)
		      (loop (+ i (string-length w))
			    (+ i (string-length w))
			    (cons (cdr s)
				  (cons (substring line istrt i) res))
			    idxs)))
		   (else (loop istrt (+ i (string-length w)) res idxs)))))
	  (else (loop istrt (+ i 1) res idxs)))))


;; Alist for argument macro definitions.
(define (scheme-args->macros args)
  (let ((m0 (string-append "@code{" (symbol->string (car args)) "}")))
    (cons 
     (cons '@arg0 m0)
     (cons
      (cons '@0 m0)
      (let recur ((i 1)
		  (args (cdr args)))
	(cond ((null? args) '())
	      ((symbol? args)
	       (let ((s (number->string i))
		     (m (string-append "@var{"
				       (symbol->string args)
				       "} @dots{}")))
		 (cons (cons (string->symbol (string-append "@" s)) m)
		       (cons (cons (string->symbol (string-append "@arg" s)) m)
			     '()))))
	      (else
	       (let ((s (number->string i))
		     (m (string-append "@var{" 
				       (symbol->string (car args))
				       "}")))
		 (cons (cons (string->symbol (string-append "@" s)) m)
		       (cons (cons (string->symbol (string-append "@arg" s)) m)
			     (recur (+ i 1) (cdr args))))))))))))

(define (schmooz-fun sexp body xdefs)
  (define (out-header sexp op)
    (let ((fun (caadr sexp))
	  (args (cdadr sexp)))
      (out 0 op fun)
      (let loop ((args args))
	(cond ((null? args))
	      ((symbol? args)
	       (out CONTLINE " " args " @dots{}"))
	      ((pair? args)
	       (out CONTLINE " " (car args))
	       (loop (cdr args)))
	      (else (slib:error sexp))))))
  (let* ((mac-list (scheme-args->macros (cadr sexp)))
	 (lines (map (lambda (bl) (substitute-macs bl mac-list))
		     body)))
    (out-header sexp "@defun ")
    (for-each (lambda (def)
		(out-header def "@defunx "))
	      xdefs)
    (for-each (lambda (line) (out 0 line)) (apply append lines))
    (out 0 "@end defun")
    (out 0)))

(define (schmooz-var sexp body xdefs)
  (let* ((name (cadr sexp))
	 (mac-list (scheme-args->macros (list name)))
	 (lines
	  (map (lambda (bl) (substitute-macs bl mac-list)) body)))
    (out 0 "@defvar " name)
    (let loop ((xdefs xdefs))
      (if (pair? xdefs)
	  (let ((sexp (car xdefs)))
	    (and (pair? sexp)
		 (eq? (car sexp) 'DEFINE)
		 (pair? (cdr sexp))
		 (symbol? (cadr sexp)))
	    (out 0 "@defvarx " (cadr sexp))
	    (loop (cdr xdefs)))))
    (for-each (lambda (line) (out 0 line)) (apply append lines))
    (out 0 "@end defvar")
    (out 0)))

;;; SCHMOOZ files.
(define (schmooz . files)
  (for-each (lambda (file)
	      (define texname (string-append (descmfilify file) ".txi"))
	      (display "Schmoozing ") (write file)
	      (display " -> ") (write texname) (newline)
	      (fluid-let ((*scheme-source* (open-file file "r"))
			  (*scheme-source-name* file)
			  (*derived-txi* (open-file texname "w")))
		(set! *output-line* 1)
		(schmooz-tops schmooz-top)
		(close-input-port *scheme-source*)
		(close-output-port *derived-txi*)))
	    files))

;;; SCHMOOZ-TOPS - schmooz top level forms.
(define (schmooz-tops schmooz-top)
  (let ((doc-lines '()))
    (define (tok1 line)
      (let loop ((i 0))
	(cond ((>= i (string-length line)) line)
	      ((or (char-whitespace? (string-ref line i))
		   (char=? #\; (string-ref line i)))
	       (substring line 0 i))
	      (else (loop (+ i 1))))))
    (define (skip-ws line)
      (do ((i 0 (+ i 1)))
	  ((or (>= i (string-length line))
	       (not (memv (string-ref line i)
			  '(#\space #\tab #\;))))
	   (substring line i (string-length line)))))
    (define (read-cmt-line)
      (cond ((eqv? #\; (peek-char *scheme-source*))
	     (read-char *scheme-source*)
	     (read-cmt-line))
	    (else (read-line *scheme-source*))))
    (define (read-newline)
      (if (char=? #\cr (read-char *scheme-source*))
	  (if (char=? #\nl (peek-char *scheme-source*))
	      (read-char *scheme-source*)
	      (report "stranded #\\cr"))))
    (define (lp c)
      (cond ((eof-object? c)
	     (cond ((pair? doc-lines)
		    (report "No definition found for @body doc lines"
			    (reverse doc-lines)))))
	    ((memv c '(#\cr #\nl))
	     (read-newline)
	     (set! *output-line* (+ 1 *output-line*))
	     (newline *derived-txi*)
	     (lp (peek-char *scheme-source*)))
	    ((char-whitespace? c)
	     (write-char (read-char *scheme-source*) *derived-txi*)
	     (lp (peek-char *scheme-source*)))
	    ((char=? c #\;)
	     (c-cmt c))
	    (else 
	     (sx))))
    (define (sx)
      (let* ((s1 (read *scheme-source*))
	     (ss (let recur ()		;Read all forms separated only by single newlines.
		   (case (peek-char *scheme-source*)
		     ((#\cr) (read-char *scheme-source*) (recur))
		     ((#\nl) (read-char *scheme-source*)
		      (if (eqv? #\( (peek-char *scheme-source*))
			  (cons (read *scheme-source*) (recur))
			  '()))
		     (else '())))))
	(cond ((eof-object? s1))
	      (else
	       (schmooz-top (if (null? ss) s1 `(DOC-BEGIN ,s1 ,@ss))
			    (reverse doc-lines))
	       (set! doc-lines '())
	       (lp (peek-char *scheme-source*))))))
    ;;Comments transcribed to generated C source files.
    (define (c-cmt c)
      (cond ((eof-object? c) (lp c))
	    ((eqv? #\; c)
	     (read-char *scheme-source*)
	     (c-cmt (peek-char *scheme-source*)))
	    ;; Escape to start Texinfo comments
	    ((eqv? #\@ c)
	     (let* ((line (read-line *scheme-source*))
		    (tok (tok1 line)))
	       (cond ((or (string=? tok "@body")
			  (string=? tok "@text"))
		      (set! doc-lines 
			    (cons (skip-ws
				   (substring line 
					      (string-length tok)
					      (string-length line)))
				  doc-lines))
		      (body-cmt (peek-char *scheme-source*)))
		     (else
		      (for-each (lambda (l) 
				  (newline *derived-txi*)
				  (display l *derived-txi*))
				(substitute-macs
				 (if (string=? tok "@")
				     (skip-ws
				      (substring line 1 (string-length line)))
				     line)
				 '()))
		      (doc-cmt (peek-char *scheme-source*))))))
	    ;; Transcribe the comment line to C source file.
	    (else
	     (read-line *scheme-source*) ;(out-c-cmt )
	     (lp (peek-char *scheme-source*)))))
    ;;Comments incorporated in generated Texinfo files.
    ;;Continue adding lines to DOC-LINES until a non-comment
    ;;line is reached (may be a blank line).
    (define (body-cmt c)
      (cond ((eof-object? c) (lp c))
	    ((eqv? #\; c)
	     (set! doc-lines (cons (read-cmt-line) doc-lines))
	     (body-cmt (peek-char *scheme-source*)))
	    ((memv c '(#\nl #\cr))
	     (read-newline)
	     (lp (peek-char *scheme-source*)))
	    ;; Allow whitespace before ; in doc comments.
	    ((char-whitespace? c)
	     (read-char *scheme-source*)
	     (body-cmt (peek-char *scheme-source*)))
	    (else 
	     (lp (peek-char *scheme-source*)))))
    ;;Comments incorporated in generated Texinfo files.
    ;;Transcribe comments to current position in Texinfo file
    ;;until a non-comment line is reached (may be a blank line).
    (define (doc-cmt c)
      (cond ((eof-object? c) (lp c))
	    ((eqv? #\; c)
	     (let* ((ls (substitute-macs (read-cmt-line) '())))
	       (for-each (lambda (l) 
			   (newline *derived-txi*)
			   (display l *derived-txi*))
			 ls)
	       ;;(out-c-cmt (car ls))
	       (doc-cmt (peek-char *scheme-source*))))
	    ((memv c '(#\nl #\cr))
	     (read-newline)
	     (newline *derived-txi*)
	     (lp (peek-char *scheme-source*)))
	    ;; Allow whitespace before ; in doc comments.
	    ((char-whitespace? c)
	     (read-char *scheme-source*)
	     (doc-cmt (peek-char *scheme-source*)))
	    (else 
	     (newline *derived-txi*)
	     (lp (peek-char *scheme-source*)))))
    (lp (peek-char *scheme-source*))))

(define (schmooz-top-doc-begin defs doc)
  (let ((s1 (car defs)))
    (cond
     ((or (not (pair? s1))
	  (not (memq (car s1) '(DEFINE DEFVAR DEFCONST))))
      (report "SCHMOOZ-TOP: no definition found for Texinfo documentation"
	      doc (car defs)))
     (else
      (let ((op1 (car s1))
	    (proc? (pair? (cadr s1))))
	(let loop ((ss (cdr defs))
		   (smatch (list s1)))
	  (if (and (pair? ss)
		   (pair? (car ss))
		   (eq? op1 (caar ss))
		   (if proc?
		       (pair? (cadar ss))
		       (not (pair? (cadar ss)))))
	      (loop (cdr ss) (cons (car ss) smatch))
	      (let ((smatch (reverse smatch)))
		(cond (proc?
		       (schmooz-fun (car smatch) doc (cdr smatch)))
		      (else
		       (schmooz-var (car smatch) doc (cdr smatch))))))))))))

;;; SCHMOOZ-TOP - schmooz top level form sexp.
(define (schmooz-top sexp . doc)
  (if (pair? doc) (set! doc (car doc)))
  (cond ((not (pair? sexp)))
	(else
	 (case (car sexp)
	   ((LOAD REQUIRE)		;If you redefine load, you lose
	    #f)
	   ((BEGIN)
	    (cond ((pair? doc)
		   (schmooz-top (cadr sexp) doc)
		   (set! doc '())
		   (for-each schmooz-top (cddr sexp)))
		  (else
		   (for-each schmooz-top (cdr sexp)))))
	   ((DOC-BEGIN)
	    (if (pair? doc)
		(begin
		  (schmooz-top-doc-begin (cdr sexp) doc)
		  (set! doc '())))
	    (for-each schmooz-top (cdr sexp)))
	   ((DEFVAR DEFINE DEFCONST)
	    (cond ((pair? (cadr sexp))
		   (set! *procedure* (caadr sexp))
		   (cond ((pair? doc)
			  (schmooz-fun sexp doc '())
			  (set! doc '()))))
		  (else
		   (cond ((pair? doc)
			  (schmooz-var sexp doc '())
			  (set! doc '())))))))
	 (or (null? doc)
	     (report 
	      "SCHMOOZ-TOP: no definition found for Texinfo documentation"
	      doc sexp))
	 (set! *procedure* #f))))
